/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <math.h>
#include "algorithms.h"
#include "graphanim.h"
#include "graph.h"
#include "treeanim.h"
#include "tree.h"

void insertionSort(GraphAnim *ga, int begin, int end, bool updatehighlight = true) {
	int *values = ga->graph->values;
	ga->highlight(begin,end);
	for (int i = begin+1; i <= end; i++) {
		if (updatehighlight) ga->highlight(begin,i);
		ga->markRed(i);
		ga->drawAnim();
		ga->moveForward(i);
		int val = values[i];
		int j = i;
		while (true) {
			if (j==begin) break;
			ga->markRed(j-1);
			ga->drawAnim();
			if (values[j-1] <= val) break;
			ga->slideRight(j-1, true);
//			values[j] = values[j-1];
			ga->clearMark(j);
			ga->drawAnim();
			j--;
		}
/*		for (j = i; j > 0 && values[j-1] > val; j--) {
			values[j] = values[j-1];
		}
*/
		if (j>begin) ga->clearMark(j-1);
		values[j] = val;
		ga->moveBackTo(i,j);
		ga->clearMark(j);
		ga->drawAnim();
	}
}

void insertionSort(GraphAnim *ga, int n) {
	insertionSort(ga,0,n-1);
}

void selectionSort(GraphAnim *ga, int begin, int end, bool updatehighlight = true) {
	int *values = ga->graph->values;
	if (!updatehighlight) ga->highlight(begin,end);
	for (int i = begin; i < end; i++) {
		if (updatehighlight) ga->highlight(i,end);
		ga->markRed(i);
		ga->drawAnim();
		ga->moveForward(i);
		int min = values[i];
		int minpos = i;
		for (int j = i+1; j <= end; j++) {
			ga->markRed(j);
			ga->drawAnim();
			if (values[j] <= min) {
				ga->moveForward(j);
				ga->moveBackTo(minpos,minpos);
				ga->clearMark(minpos);
				min = values[j];
				minpos = j;
			} else {
				ga->clearMark(j);
			}
		}
		ga->drawAnim();
		if (minpos!=i) {
			ga->moveTo(i,minpos);
			ga->moveBackTo(minpos,i);
			values[i] = min;
		} else {
			ga->moveBackTo(minpos,minpos);
		}
		ga->clearMark(i);
		ga->drawAnim();
	}
}

void selectionSort(GraphAnim *ga, int n) {
	selectionSort(ga,0,n-1);
}

void swap(GraphAnim *ga, int *values, int a, int b) {
//	if (a==b) return;
	ga->animSwap(a,b,true);
//	int tmp = values[a];
//	values[a] = values[b];
//	values[b] = tmp;
}

int findMedianThree(GraphAnim *ga, int *values, int begin, int end) {
	int med = (begin+end)/2;
	ga->markRed(begin);
	ga->markRed(med);
	ga->drawAnim();
	if (values[begin]>values[med]) swap(ga,values,begin,med);
	ga->clearMark(med);
	ga->markRed(end);
	ga->drawAnim();
	if (values[begin]>values[end]) swap(ga,values,begin,end);
	ga->clearMark(begin);
	ga->markRed(med);
	ga->drawAnim();
	if (values[med]>values[end]) swap(ga,values,med,end);
	ga->clearMarks();
	ga->drawAnim();
	swap(ga,values,med,end-1);
	return values[end-1];
}

void quickSort(GraphAnim *ga, int begin, int end, bool cutoff = false) {
	int *values = ga->graph->values;
	if (cutoff && end-begin+1<=10) {
		insertionSort(ga,begin,end,false);
//		selectionSort(ga,begin,end,false);
		return;
	}
	ga->highlight(begin,end);
	if (end<=begin) return;
	if (end-begin+1==2) {
		ga->markRed(begin);
		ga->markRed(end);
		ga->drawAnim();
		if (values[begin]>values[end]) swap(ga,values,begin,end);
		ga->clearMarks();
		ga->drawAnim();
		return;
	}
	int med = findMedianThree(ga,values,begin,end);
	if (end-begin+1<=3) return;
	int left = begin;
	int right = end-1;
	ga->markRed(end-1);
	ga->drawAnim();
	while (true) {
//		while (values[++left]<med);
		while (true) {
			left++;
			ga->markRed(left);
			ga->drawAnim();
			if (values[left]>=med) break;
			ga->clearMark(left);
		}
//		while (values[--right]>med);
		while (true) {
			right--;
			ga->markRed(right);
			ga->drawAnim();
			if (values[right]<=med) break;
			if (right!=left)
				ga->clearMark(right);
		}
		if (left < right) {
			swap(ga,values,left,right);
			ga->clearMark(left);
			ga->clearMark(right);
			ga->drawAnim();
		} else {
			ga->clearAnims();
//			ga->clearMark(left);
			ga->clearMark(right);
			ga->markRed(left);
			ga->drawAnim();
			break;
		}
	}
	swap(ga,values,left,end-1);
	ga->clearMarks();
	ga->drawAnim();
	quickSort(ga,begin,right,cutoff);
	quickSort(ga,right+2,end,cutoff);
}



void shellSort(GraphAnim *ga, int n) {
	int *values = ga->graph->values;
//	ga->highlight(0,n-1);
	int gap = int(log(n)/log(2));
	gap = int(pow(2,gap))-1;
	for (; gap > 0; gap = (gap+1)/2-1) {
		if (gap==1) ga->markGreen(0);
		for (int i = gap; i < n; i++) {
			if (gap>1) {
				for (int j = i; j >= 0; j -= gap)
					ga->markGreen(j);
			} else ga->markGreen(i);
			ga->drawAnim();
			ga->markRed(i);
			ga->drawAnim();
			ga->moveForward(i);
			int val = values[i];
			int j = i;
			while (true) {
				if (j<gap) break;
				ga->markRed(j-gap);
				ga->drawAnim();
				if (values[j-gap] <= val) break;
				ga->moveTo(j-gap,j,true);
//				values[j] = values[j-gap];
//				ga->clearMark(j);
				ga->markGreen(j);
				ga->drawAnim();
				j -= gap;
			}
//			if (j>=gap) ga->clearMark(j-gap);
			if (j>=gap) ga->markGreen(j-gap);
/*			for (; j >= gap && values[j-gap]>val; j -= gap)
				values[j] = values[j-gap];
*/
			values[j] = val;
			ga->moveBackTo(i,j);
//			ga->clearMark(j);
			ga->markGreen(j);
			ga->drawAnim();
			if (gap>1) {
				ga->clearMarks();
				ga->drawAnim();
			}
		}
	}
	ga->clearMarks();
	ga->drawAnim();
}

void bubbleSort(GraphAnim *ga, int n) {
	int *values = ga->graph->values;
	ga->highlight(0,n-1);
	for (int i = 0; i < n-1; i++) {
		ga->highlight(0,n-1-i);
		ga->markRed(0);
		bool swaps = false;
		for (int j = 0; j < n-1-i; j++) {
			ga->markRed(j+1);
			ga->drawAnim();
			if (values[j]>values[j+1]) {
//				swap(ga,values,j,j+1);
				ga->swapNeighbors(j);
				swaps = true;
			}
			ga->clearMark(j);
		}
		ga->clearMark(n-1-i);
		ga->drawAnim();
		if (!swaps) break;
	}
}

void biDirBubbleSort(GraphAnim *ga, int n) {
	int *values = ga->graph->values;
	int left = 0;
	int right = n-1;
	while (left<right) {
		ga->highlight(left,right);
		ga->markRed(left);
		bool swaps = false;
		for (int j = left; j < right; j++) {
			ga->markRed(j+1);
			ga->drawAnim();
			if (values[j]>values[j+1]) {
				ga->swapNeighbors(j);
				swaps = true;
			}
			ga->clearMark(j);
		}
		ga->clearMark(right);
		ga->drawAnim();
		right--;
		if (!swaps) break;
		swaps = false;
		ga->highlight(left,right);
		ga->markRed(right);
		for (int j = right; j > left; j--) {
			ga->markRed(j-1);
			ga->drawAnim();
			if (values[j-1]>values[j]) {
				ga->swapNeighbors(j-1);
				swaps = true;
			}
			ga->clearMark(j);
		}
		ga->clearMark(left);
		ga->drawAnim();
		left++;
		if (!swaps) break;
	}
}


void merge(GraphAnim *ga, int left, int leftend, int right, int rightend) {
	ga->highlight(left,rightend);
	int *values = ga->graph->values;
	ga->markRed(left);
	ga->markRed(right);
	ga->drawAnim();
	int out = ga->graph->numValues/2;
	int begin = left;
	while (true) {
		if (values[left]<values[right]) {
//			newvalues[out++] = values[left++];
			ga->moveTo(left,out,true);
			ga->clearMark(out);
			left++;
			out++;
			if (left>leftend) break;
			ga->markRed(left);
		} else {
//			newvalues[out++] = values[right++];
			ga->moveTo(right,out,true);
			ga->clearMark(out);
			right++;
			out++;
			if (right>rightend) break;
			ga->markRed(right);
		}
		ga->drawAnim();
	}
	if (left<=leftend) ga->clearMark(left);
	if (right<=rightend) ga->clearMark(right);
	ga->drawAnim();
	while (left<=leftend) {
//		newvalues[out++] = values[left++];
		ga->moveTo(left,out,true);
		left++;
		out++;
	}
	while (right<=rightend) {
//		newvalues[out++] = values[right++];
		ga->moveTo(right,out,true);
		right++;
		out++;
	}
	out = ga->graph->numValues/2;
	while (begin<=rightend) {
//		values[begin++] = newvalues[out++];
		ga->moveTo(out,begin,true);
		out++;
		begin++;
	}
}

void mergeRecursive(GraphAnim *ga, int begin, int end) {
	if (end<=begin) return;
	if (end-begin>1) ga->highlight(begin,end);
	int middle = (begin+end)/2;
	mergeRecursive(ga,begin,middle);
	mergeRecursive(ga,middle+1,end);
	merge(ga,begin,middle,middle+1,end);
}

void mergeSort(GraphAnim *ga, int n) {
	ga->graph->setZoomHighlight(false);
	ga->highlight(0,n-1);
	ga->clearZoom();
	ga->zoom(0,n-1);
	ga->graph->resize(n*2);
	ga->zoom(0,n*2-1);

	mergeRecursive(ga,0,n-1);

	ga->graph->resize(n);
	ga->zoom(0,n-1);
}

void merge2(GraphAnim *ga, int left, int leftend, int right, int rightend, int buffer) {
	ga->highlight(left,rightend);

	int i = left;
	int out = buffer;
	while (i<=leftend) {
		ga->moveTo(i,out,true);
		i++;
		out++;
	}

	int *values = ga->graph->values;
	out = left;
	leftend = leftend-left+buffer;
	left = buffer;
	ga->markRed(left);
	ga->markRed(right);
	ga->drawAnim();
	while (true) {
		if (values[left]<values[right]) {
//			newvalues[out++] = values[left++];
			ga->moveTo(left,out,true);
			ga->clearMark(out);
			left++;
			out++;
			if (left>leftend) break;
			ga->markRed(left);
		} else {
//			newvalues[out++] = values[right++];
			ga->moveTo(right,out,true);
			ga->clearMark(out);
			right++;
			out++;
			if (right>rightend) break;
			ga->markRed(right);
		}
		ga->drawAnim();
	}
	if (left<=leftend) ga->clearMark(left);
	if (right<=rightend) ga->clearMark(right);
	ga->drawAnim();
	while (left<=leftend) {
//		newvalues[out++] = values[left++];
		ga->moveTo(left,out,true);
		left++;
		out++;
	}
}

void mergeRecursive2(GraphAnim *ga, int begin, int end, int buffer) {
	if (end<=begin) return;
	if (end-begin>1) ga->highlight(begin,end);
	int middle = (begin+end)/2;
	mergeRecursive2(ga,begin,middle,buffer);
	mergeRecursive2(ga,middle+1,end,buffer);
	merge2(ga,begin,middle,middle+1,end,buffer);
}

void mergeSort2(GraphAnim *ga, int n) {
	ga->graph->setZoomHighlight(false);
	ga->highlight(0,n-1);
	ga->clearZoom();
	ga->zoom(0,n-1);
	ga->graph->resize(n+n/2);
	ga->zoom(0,n+n/2-1);

	mergeRecursive2(ga,0,n-1,n);

	ga->graph->resize(n);
	ga->zoom(0,n-1);
}

void percolateDown(GraphAnim *ga, int i, int max) {
	int *values = ga->graph->values;
	ga->markRed(i);
	ga->drawAnim();
	ga->moveForward(i);
	int orig = i;
	int origval = values[i];
	while (true) {
		int child = 2*i+1; // 2*(i+1)-1 = 2*i+1
		if (child>=max) break;
		ga->markRed(child);
		if (child!=max-1) {
			ga->markRed(child+1);
			ga->drawAnim();
			if (values[child+1]>values[child]) {
				ga->clearMark(child);
				child++;
			} else {
				ga->clearMark(child+1);
			}
		}
		ga->drawAnim();
//		if (child!=max-1 && values[child+1]>values[child])
//			child++;
		if (origval<values[child]) {
			ga->moveTo(child,i,true);
			ga->clearMark(i);
			ga->drawAnim();
//			swap(ga,values,i,child);
			i = child;
		} else {
			ga->clearMark(child);
			ga->drawAnim();
			break;
		}
	}
	values[i] = origval;
	ga->moveBackTo(orig,i);
	ga->clearMark(i);
	ga->drawAnim();
}

void heapSort(GraphAnim *ga, int n) {
	ga->highlight(0,n-1);
	for (int i = n/2-1; i >= 0; i--) {
		percolateDown(ga,i,n);
	}
	for (int i = n-1; i > 0; i--) {
		swap(ga,ga->graph->values,0,i);
		ga->highlight(0,i-1);
		percolateDown(ga,0,i);
	}
}

#define QUIT_INTERRUPTED(anim)			\
	if (setjmp(*(anim)->jmpEnv())) {	\
		return 0;			\
	}

int quickSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	quickSort(ga,0,ga->graph->numValues-1);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int quickSort2Thread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	quickSort(ga,0,ga->graph->numValues-1,true);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int insertionSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	insertionSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int bubbleSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	bubbleSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int biDirBubbleSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	biDirBubbleSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int shellSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	shellSort(ga,ga->graph->numValues);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int mergeSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	mergeSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int mergeSort2Thread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	mergeSort2(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int heapSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	heapSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

int selectionSortThread(void *arg) {
	GraphAnim *ga = (GraphAnim*) arg;
	QUIT_INTERRUPTED(ga);
	ga->clearHighlight();
	selectionSort(ga,ga->graph->numValues);
	ga->highlight(0,ga->graph->numValues-1);
	ga->drawAnim(1000);
	ga->setDone(true);
	return 0;
}

void addBalanced(TreeAnim *ta, int treenum, int *values, int left, int right) {
	if (right<left) return;
	int m = (left+right)/2;
	ta->add(treenum,values[m]);
	addBalanced(ta,treenum,values,left,m-1);
	addBalanced(ta,treenum,values,m+1,right);
}

int intCompare(const void *p1, const void *p2) {
	int *i1 = (int*) p1;
	int *i2 = (int*) p2;
	if (*i1<*i2) return -1;
	else if (*i1>*i2) return 1;
	return 0;
}

int addBalancedThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	qsort(ta->addValues,ta->maxval,sizeof(int),intCompare);
	addBalanced(ta,treenum,ta->addValues,0,ta->maxval-1);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->remove(treenum,ta->remValues[i]);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int bstThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	for (int i = 0; i < ta->maxval; i++)
		ta->add(treenum,ta->addValues[i]);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->remove(treenum,ta->remValues[i]);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int avlThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	for (int i = 0; i < ta->maxval; i++)
		ta->addAvl(treenum,ta->addValues[i]);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->removeAvl(treenum,ta->remValues[i]);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int splayThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	for (int i = 0; i < ta->maxval; i++)
		ta->addSplay(treenum,ta->addValues[i]);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->removeSplay(treenum,ta->remValues[i]);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int heapThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	for (int i = 0; i < ta->maxval; i++)
		ta->addHeap(treenum,ta->addValues[i]);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->removeHeap(treenum);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int buildHeapThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	ta->buildHeap(treenum,ta->addValues,ta->maxval);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->removeHeap(treenum);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}

int redBlackThread(void *arg) {
	TreeAnim *ta = (TreeAnim*) arg;
	QUIT_INTERRUPTED(ta);
	ta->inserted = false;
	ta->allInserted = false;

	ta->forest->maxval = ta->maxval;

	int treenum = ta->addTree();

	for (int i = 0; i < ta->maxval; i++)
		ta->addRedBlack(treenum,ta->addValues[i]);

	ta->drawAnim(1000);
	ta->doneInserting();

	for (int i = 0; i < ta->maxval; i++)
		ta->removeRedBlack(treenum,ta->remValues[i]);

	ta->forest->clear();

	ta->drawAnim(1000);
	ta->setDone(true);
	return 0;
}


struct algorithm sortalgorithms[] = {
//	{ "quicksort2", quickSort2Thread },
	{ "quicksort", "Quicksort", quickSortThread },
	{ "mergesort2", "Optimized merge sort", mergeSort2Thread },
	{ "mergesort", "Merge sort", mergeSortThread },
	{ "heapsort", "Heap sort", heapSortThread },
	{ "shellsort", "Shell sort", shellSortThread },
	{ "selectionsort", "Selection sort", selectionSortThread },
	{ "insertionsort", "Insertion sort", insertionSortThread },
	{ "bidirbubblesort", "Bi-directional bubble sort", biDirBubbleSortThread },
	{ "bubblesort", "Bubble sort", bubbleSortThread },
	{ NULL, NULL, NULL }
};

struct OrderDesc orders[] = {
	{ "random", "Random", RANDOM },
	{ "sorted", "Sorted", SORTED },
	{ "reverse", "Reverse", REVERSE },
	{ "semisorted", "Semisorted", SEMISORTED },
	{ NULL, NULL, RANDOM }
};

struct algorithm treealgorithms[] = {
	{ "bst", "Plain BST", bstThread },
	{ "addbalanced", "BST with presorting of input", addBalancedThread },
	{ "avl", "AVL tree", avlThread },
	{ "splay", "Bottom-up splay tree", splayThread },
	{ "heap", "Binary heap", heapThread },
	{ "buildheap", "Build binary heap", buildHeapThread },
	{ "redblack", "Red-black tree", redBlackThread },
	{ NULL, NULL, NULL }
};

