/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __ALGORITHMS_H
#define __ALGORITHMS_H

typedef	int (*threadFunc)(void *);
struct algorithm {
	const char *name;
	const char *description;
	threadFunc func;
};

extern struct algorithm sortalgorithms[];
extern struct algorithm treealgorithms[];

enum DataOrder { RANDOM, SORTED, REVERSE, SEMISORTED };

struct OrderDesc {
	const char *name;
	const char *description;
	DataOrder order;
};

extern struct OrderDesc orders[];

#include <setjmp.h>
#include <stdio.h>
#include "cothread.h"

class Anim {
private:
	cothread thread;
	bool done;
	jmp_buf jmpenv;
	const struct algorithm* alg;
protected:
	int animtime;

public:
	Anim() {
		animtime = 0;
		done = true;
		thread = NULL;
	}
	virtual ~Anim() {
	}
	enum AnimType { GRAPH, TREE };
	virtual void setTime(int time) { animtime = time; }
	virtual void addTime(int time) { animtime += time; }
	virtual bool isDone() { return done; }
	virtual void setDone(bool d = true) { done = d; }
	virtual void drawAnim(int msec) = 0;
	virtual bool updateAnims() = 0;
	virtual void clearAnims() = 0;
	virtual AnimType getType() = 0;
	virtual void draw(int time, int difftime) = 0;
	const struct algorithm* getAlgorithm() const { return alg; }
	void setAlgorithm(const struct algorithm* a) { alg = a; }
	jmp_buf *jmpEnv() { return &jmpenv; }
	cothread getThread() const { return thread; }
	void setThread(cothread t) { thread = t; }

	void swapThreads() {
		cothread_swap(thread);
		if (done && cothread_active(thread)) longjmp(jmpenv, 1);
	}


};


#endif
