/*
    Cothread
    Copyright 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include "cothread.h"

#ifndef COTHREAD_ASM_SUPPORTED

#include <SDL.h>
#include <SDL_thread.h>
#include <stdlib.h>

struct _cothread {
	int current, done, join;
	cothread_func func;
	void* arg;
	SDL_sem *sem[2];
	SDL_Thread* thread;
};

void cothread_start(cothread thread);
void cothread_end(cothread thread);

static int thread_init(void* arg) {
	int result;
	cothread thread = (cothread) arg;

	cothread_start(thread);
	result = thread->func(thread->arg);
	cothread_end(thread);

	return result;
}

cothread cothread_create(cothread_func func, void *arg) {
	cothread thread = (cothread) malloc(sizeof(struct _cothread));

	thread->sem[0] = SDL_CreateSemaphore(0);
	thread->sem[1] = SDL_CreateSemaphore(0);
	thread->current = 0;
	thread->done = 0;
	thread->join = 0;

	thread->func = func;
	thread->arg = arg;
	thread->thread = SDL_CreateThread(thread_init, thread);

	return thread;
}

void cothread_swap(cothread thread) {
	int next, curr;

	if (thread->current == 0 && thread->done) return;
	if (thread->current == 1 && thread->join) return;

	curr = thread->current;
	next = (thread->current+1)&1;
	thread->current = next;

	SDL_SemPost(thread->sem[next]);
	SDL_SemWait(thread->sem[curr]);
}

void cothread_start(cothread thread) {
	SDL_SemWait(thread->sem[1]);
}

void cothread_end(cothread thread) {
	thread->done = 1;
	thread->current = 0;
	SDL_SemPost(thread->sem[0]);
}

int cothread_join(cothread thread) {
	int result;

	thread->join = 1;
	thread->current = 1;
	SDL_SemPost(thread->sem[1]);
	SDL_WaitThread(thread->thread, &result);

	SDL_DestroySemaphore(thread->sem[0]);
	SDL_DestroySemaphore(thread->sem[1]);
	free(thread);

	return result;
}

int cothread_running(cothread thread) {
	return !thread->done;
}

int cothread_active(cothread thread) {
	return thread->current;
}

#endif

