/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __GRAPH_H
#define __GRAPH_H


#include <SDL.h>
#include <SDL_opengl.h>
#include <vector>

using std::vector;

#include "algorithms.h" // for DataOrder


class Movable {
public:
	int id;
	GLfloat pos[3];
	GLfloat size[3];
	GLfloat color[3];
};

class Graph {
private:
	vector<Movable> movables;
	GLfloat highlightPos[3];
	GLfloat highlightSize[3];
	GLfloat zoomPos[3];
	GLfloat zoomSize[3];


	GLfloat *colors;
	GLfloat defaultColor[3];
	vector<bool> hidden;
	bool zoomHighlight;

	int max;

public:
	void setColor(int n, GLfloat *c);
	void getColor(int n, GLfloat *c);
	void getDefaultColor(GLfloat *c);
	void resetColor(int n);
//	bool hasMark(int n);

	void hide(int n, bool h);
	void setMovable(int id, GLfloat *pos, GLfloat *size, GLfloat *color);
	void getMovable(int id, GLfloat *pos, GLfloat *size, GLfloat *color);
	void removeMovable(int id);

	void setHighlight(GLfloat *pos, GLfloat *size);
	void getHighlight(GLfloat *pos, GLfloat *size);
	void removeHighlight();
	void setZoom(GLfloat *pos, GLfloat *size);
	void getZoom(GLfloat *pos, GLfloat *size);
	void removeZoom();


	void setZoomHighlight(bool z);

	int *values;
	int numValues;

	Graph(int n = 60);
	~Graph();
	void initValues(DataOrder order);
	void copyValues(Graph *g);
	void resize(int n);
	void setMax(int m);

	void getPos(int i, GLfloat *pos);
	void getSize(int i, GLfloat *size);
	void draw(int time, int difftime);
};







#endif
