/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __SDLTTF_H
#define __SDLTTF_H

#ifdef HAVE_SDL_TTF

#include <SDL.h>
#include <SDL_ttf.h>

class Font {
private:
	GLuint lists;
	int offset, max;
	GLuint textures[256];
	int numtextures;
	int advance[256];
	int height[256];
	int heightA;

	int curx, cury, nexty;
	SDL_Surface *tmpsurf;

	void newTexture();
	void initSurf();
	void uploadSurf();
	void addGlyph(TTF_Font *font, int c);
	void getSize(const char *str, int *width, int *height);
	void drawOverlay2D(float height, float x, float y, char *buffer);
public:
	Font(const char *file, int size, int first = 32, int last = 255);
	~Font();
	void print(float height, const char *fmt, ...);
	void printCentered(float height, const char *fmt, ...);
	void printCentered2D(float height, const char *fmt, ...);
	void printOverlay2D(float height, float x, float y, const char *fmt, ...);
	void printOverlay2D(int height, int x, int y, const char *fmt, ...);
};

#else

class Font {
public:
	Font(const char *file, int size, int first = 32, int last = 255) {}
	~Font() {}
	void print(float height, const char *fmt, ...) {}
	void printCentered(float height, const char *fmt, ...) {}
	void printCentered2D(float height, const char *fmt, ...) {}
	void printOverlay2D(float height, float x, float y, const char *fmt, ...) {}
	void printOverlay2D(int height, int x, int y, const char *fmt, ...) {}
};

#endif

#endif
