/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __SPLINE_H
#define __SPLINE_H

class Spline {
private:
	GLfloat start[3], end[3], waypoint1[3], waypoint2[3], waypoint3[3];
	float divpos;
	float len1, len2, len3, len4;
	float totlen;
	static float getLength(const GLfloat *v1, const GLfloat *v2, const GLfloat *v3, float *lp1, float *lp2);
public:
	Spline(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3);
	Spline();
	~Spline();
	void init(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3);
	float getLength();
	void calc(float pos, GLfloat *out);
};

class Spline3 {
private:
	GLfloat start[3], end[3], waypoint1[3], waypoint2[3];
	float length;
public:
	Spline3(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3);
	Spline3();
	~Spline3();
	void init(const GLfloat *v0, const GLfloat *v1, const GLfloat *v2, const GLfloat *v3);
	float getLength();
	void calc(float pos, GLfloat *out);
};

class Rotation {
private:
	Spline3 posSpline, spline1, spline2;
	int ax1, ax2, ax3;
public:
	Rotation(const GLfloat *m0, const GLfloat *m1, const GLfloat *m2, const GLfloat *m3, int a = 0);
	Rotation(const GLfloat *m1, const GLfloat *m2, int a = 0);
	Rotation();
	void init(const GLfloat *m0, const GLfloat *m1, const GLfloat *m2, const GLfloat *m3, int a = 0);
	void init(const GLfloat *m1, const GLfloat *m2, int a = 0);
	~Rotation();
	void rotate(float pos, GLfloat *out);
};

#endif
