/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __TREE_H
#define __TREE_H

#include <SDL.h>
#include <SDL_opengl.h>
#include <vector>

using std::vector;


#define NODEWIDTH 1.0
#define NODEHEIGHT 1.4
#define TREESPACING 1.0


void reinitQuadrics();

class Node {
public:
	int val;
	bool spacing;
	bool hide;
	Node *left, *right;
	float width, height;
	int iheight, balance;
	bool red;
	int maxval;
	float parentlength;
	GLfloat color[3];
	bool defaultColor;

	float updateWidth();
	float updateHeight();

	Node(int v);
	Node(int v, float width, float height);
	Node(float width, float height);
	~Node();

	void draw(GLfloat *parent);
	void drawCylinder(GLfloat *end, float radius);
	void drawNode();
	void setMaxval(int max);

	void addPos(Node *n, GLfloat *pos);
	void setDefaultColor(GLfloat *color);
};

class Tree {
public:
	Node *root;
	GLfloat pos[3];
	enum PosType { TYPE_ABSOLUTE, TYPE_RELATIVE };
	PosType type;
	int reltree;
	Node *relnode;
	vector<Node*> heap;

	Tree();
	~Tree();
	void add(int i);
	void add(int i, Node *newnode);
	Node **getParentLink(int i);
	void addPos(Node *n, GLfloat *pos);

	Node **getHeapLink(int pos);
	void addHeapPos(int pos, GLfloat *pos);
	void addHeapSubTreePos(int pos, GLfloat *pos);

};

class Forest {
private:
	void calcView(GLfloat *upleft, GLfloat *upright, GLfloat *downleft);
	GLfloat defaultColor[3];
	GLfloat modelview[16];
	GLint viewport[4];

public:
	vector<Tree*> trees;
	vector<Tree*> movables;
	int maxval;

	Forest();
	~Forest();
	void clear();

	void setMinSize(GLfloat *upleft, GLfloat *upright, GLfloat *downleft);
	void draw(int time, int difftime);
	void getTreePos(int tree, GLfloat *pos);
	void getTreePos(Tree *t, GLfloat *pos);
	void getSubTreePos(int tree, Node *n, GLfloat *pos);
	void getSubTreePos(Tree *t, Node *n, GLfloat *pos);
	void getNodePos(int tree, Node *n, GLfloat *pos);
	void getNodePos(Tree *t, Node *n, GLfloat *pos);
	float getTopY();
	float getBottomY();
	void getDefaultColor(GLfloat *color);
};


#endif
