/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __TREEANIM_H
#define __TREEANIM_H


#include <vector>

using std::vector;

#include "algorithms.h" // for Anim


class Forest;
class Node;
class Tree;

class TreeAnimItem;

class TreeAnim : public Anim {
private:
	vector<TreeAnimItem> anims;
	bool updateAnim(TreeAnimItem &a);
public:
	TreeAnim(Forest *f);
	~TreeAnim();

	int maxval;
	int *addValues;
	int *remValues;
	bool inserted;
	bool allInserted;

	void drawAnim(int msec = 0);
	bool updateAnims();
	void clearAnims();
	AnimType getType() { return TREE; }
	void draw(int time, int difftime);

	Forest *forest;
	void scaleSpacing(Node *n, float w, float h, bool draw = true, int offset = 0);
	void dropToTree(Tree *t, int treenum, float offsetX = 0, float offsetY = 0);
	void dropDown(Tree *t, int treenum, float offsetX = 0, float offsetY = 0, bool haspos = false);
	void jumpIntoTree(Tree *t, int treenum);
	void growParent(Node *n, bool draw = true);
	void removeParent(Node *n, bool draw = true);
	void setRelativeToRoot(Tree *t, int treenum);
	void setRelativeToNode(Tree *t, int treenum, Node *n);
	void setAbsolute(Tree *t);
	void fadeColor(Node *n, GLfloat *start, GLfloat *end);
	void markColor(Node *n, GLfloat *color);
	void markHalfColor(Node *n, GLfloat *color);
	void setColor(Node *n, GLfloat *color);
	void setHalfColor(Node *n, GLfloat *color);
	void markRed(Node *n);
	void setRed(Node *n);
	void markBlack(Node *n);
	void setBlack(Node *n);
	void markGreen(Node *n);
	void markGreenBlack(Node *n);
	void markGreenRed(Node *n);
	void setGreen(Node *n);
	void clearMark(Node *n);
	void moveTree(Tree *t, GLfloat *endpos, bool draw = true);
	void jumpTree(Tree *t, GLfloat *endpos, bool draw = true);

	void shuffle(int *values, int max);
	void slightShuffle(int *values, int max);
	void initValues(DataOrder order, int n = 25);
	void copyValues(TreeAnim *ta);
	void doneInserting();


	// treeanimbst.cpp
	void add(int tree, int val, bool clear = true);
	void remove(int tree, int val);
	void removeLeaf(Node **removepos, int treenum);
	void removeSingleChildNode(Node **removepos, Node *parent, int treenum);
	void removeSubRoot(Node **removepos, int treenum);
	void removeSimpleSubRoot(Node **removepos, int treenum);
	void copyRemove(Node **removepos, int treenum, Node *replacement, Node *parent);
	int addTree();

	// treeanimavl.cpp
	void rotateLeft(Tree *t, Node **rootpos, bool clear = true);
	void rotateRight(Tree *t, Node **rootpos, bool clear = true);
	bool balanceAvl(int treenum, Node **cur, int val, bool quit, bool skipnode);
	void addAvl(int tree, int val);
	void removeAvl(int tree, int val);

	// treeanimsplay.cpp
	float scaleSpacingRotLeft(Node *spacing, Node *n, Tree *t, float minheight);
	float scaleSpacingRotRight(Node *spacing, Node *n, Tree *t, float minheight);
	void splay(Tree *t, Node **root, int val, bool mark, Node *spacing = NULL, float minheight = 0);
	void addSplay(int tree, int val);
	void removeSplay(int tree, int val);

	// treeanimheap.cpp
	void percolateUp(Tree *tree, Tree *t, int npos, GLfloat *lastpos);
	void percolateDown(Tree *tree, Tree *t, int npos, GLfloat *lastpos, Tree *l, Tree *r);
	void addHeap(int tree, int val);
	void removeHeap(int tree);
	void buildHeap(int tree, int *values, int nvalues);

	// treeanimredblack.cpp
	void reorient(Tree *t, Node *current, Node *parent, Node *grand, Node **grandpos);
	void addRedBlack(int tree, int val);
	void removeRedBlack(int tree, int val);

};



#endif
