/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include "treeanim.h"
#include "tree.h"

void TreeAnim::rotateLeft(Tree *t, Node **rootpos, bool clear) {
	GLfloat rpos[3], nrpos[3], cpos[3];
	Node *root = *rootpos;
	Node *nroot = root->right;
	Node *child = nroot->left;
	forest->getSubTreePos(t,root,rpos);
	forest->getNodePos(t,nroot,nrpos);
	if (child) forest->getSubTreePos(t,child,cpos);
	removeParent(root,false);
	removeParent(nroot,false);
	if (child) removeParent(child,false);
	drawAnim();

	Tree *rtree = new Tree();
	rtree->root = root;
	*rootpos = new Node(root->width,root->height);
	memcpy(rtree->pos,rpos,sizeof(GLfloat)*3);
	forest->movables.push_back(rtree);

	Tree *nrtree = new Tree();
	nrtree->root = nroot;
	root->right = NULL;
	memcpy(nrtree->pos,nrpos,sizeof(GLfloat)*3);
	forest->movables.push_back(nrtree);

	Tree *ctree;
	if (child) {
		ctree = new Tree();
		ctree->root = child;
		nroot->left = NULL;
		memcpy(ctree->pos,cpos,sizeof(GLfloat)*3);
		forest->movables.push_back(ctree);
	}

	rpos[1] -= NODEHEIGHT;
	nrpos[1] += NODEHEIGHT;
	moveTree(rtree,rpos,false);
	moveTree(nrtree,nrpos,false);

	float nheight = -NODEHEIGHT;
	if (nroot->right) nheight = nroot->right->height;
	float rheight = -NODEHEIGHT;
	if (root->left) rheight = root->left->height;
	if (child)
		if (child->height>rheight) rheight = child->height;
	rheight += NODEHEIGHT;
	if (rheight>nheight) nheight = rheight;
	nheight += NODEHEIGHT;
	scaleSpacing(*rootpos,(*rootpos)->width,nheight,false);
	drawAnim();

	delete *rootpos;
	*rootpos = nroot;
	nrtree->root = NULL;
	delete nrtree;
	nroot->left = root;
	rtree->root = NULL;
	delete rtree;
	root->right = child;
	if (child) {
		ctree->root = NULL;
		delete ctree;
		forest->movables.pop_back();
	}
	forest->movables.pop_back();
	forest->movables.pop_back();
	growParent(nroot,false);
	growParent(root,false);
	if (clear) clearMark(root);
	if (child) growParent(child,false);
	drawAnim();
}

void TreeAnim::rotateRight(Tree *t, Node **rootpos, bool clear) {
	GLfloat rpos[3], nrpos[3], cpos[3];
	Node *root = *rootpos;
	Node *nroot = root->left;
	Node *child = nroot->right;
	forest->getNodePos(t,root,rpos);
	forest->getSubTreePos(t,nroot,nrpos);
	if (child) forest->getSubTreePos(t,child,cpos);
	removeParent(root,false);
	removeParent(nroot,false);
	if (child) removeParent(child,false);
	drawAnim();

	Tree *rtree = new Tree();
	rtree->root = root;
	*rootpos = new Node(root->width,root->height);
	memcpy(rtree->pos,rpos,sizeof(GLfloat)*3);
	forest->movables.push_back(rtree);

	Tree *nrtree = new Tree();
	nrtree->root = nroot;
	root->left = NULL;
	memcpy(nrtree->pos,nrpos,sizeof(GLfloat)*3);
	forest->movables.push_back(nrtree);

	Tree *ctree;
	if (child) {
		ctree = new Tree();
		ctree->root = child;
		nroot->right = NULL;
		memcpy(ctree->pos,cpos,sizeof(GLfloat)*3);
		forest->movables.push_back(ctree);
	}

	rpos[1] -= NODEHEIGHT;
	nrpos[1] += NODEHEIGHT;
	moveTree(rtree,rpos,false);
	moveTree(nrtree,nrpos,false);

	float nheight = -NODEHEIGHT;
	if (nroot->left) nheight = nroot->left->height;
	float rheight = -NODEHEIGHT;
	if (root->right) rheight = root->right->height;
	if (child)
		if (child->height>rheight) rheight = child->height;
	rheight += NODEHEIGHT;
	if (rheight>nheight) nheight = rheight;
	nheight += NODEHEIGHT;
	scaleSpacing(*rootpos,(*rootpos)->width,nheight,false);
	drawAnim();

	delete *rootpos;
	*rootpos = nroot;
	nrtree->root = NULL;
	delete nrtree;
	nroot->right = root;
	rtree->root = NULL;
	delete rtree;
	root->left = child;
	if (child) {
		ctree->root = NULL;
		delete ctree;
		forest->movables.pop_back();
	}
	forest->movables.pop_back();
	forest->movables.pop_back();
	growParent(nroot,false);
	growParent(root,false);
	if (clear) clearMark(root);
	if (child) growParent(child,false);
	drawAnim();
}

bool TreeAnim::balanceAvl(int treenum, Node **cur, int val, bool quit, bool skipnode) {
	if (cur==NULL) return false;
	Tree *tree = forest->trees[treenum];
	Node *n = *cur;
	if (n==NULL) return false;
	if (skipnode && n->val==val) return false;
	bool parent = false;
	if (n->val>val) {
		if (n->left) {
			if (n->left->val==val) parent = true;
			if (balanceAvl(treenum,&n->left,val,quit,skipnode)) return true;
		}
	} else if (n->val<val) {
		if (n->right) {
			if (n->right->val==val) parent = true;
			if (balanceAvl(treenum,&n->right,val,quit,skipnode)) return true;
		}
	}
	if (!parent || !skipnode) markRed(n);
	drawAnim();
	if (n->balance>1) {
		if (n->left->balance>0) {
			rotateRight(tree,cur);
		} else {
			rotateLeft(tree,&n->left);
			rotateRight(tree,cur);
		}
		tree->root->updateHeight();
		return quit;
	} else if (n->balance<-1) {
		if (n->right->balance<0) {
			rotateLeft(tree,cur);
		} else {
			rotateRight(tree,&n->right);
			rotateLeft(tree,cur);
		}
		tree->root->updateHeight();
		return quit;
	}
	clearMark(n);
	return false;
}

void TreeAnim::addAvl(int treenum, int val) {
	add(treenum,val,false);
	Tree *tree = forest->trees[treenum];
	tree->root->updateHeight();
	balanceAvl(treenum,&tree->root,val,true,true);
	drawAnim();
}

void TreeAnim::removeAvl(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	Node *n = tree->root;
	Node *parent = NULL;
	while (true) {
		if (val<n->val) {
			parent = n;
			n = n->left;
		} else if (val>n->val) {
			parent = n;
			n = n->right;
		} else {
			break;
		}
	}
	int balanceval = val;
	bool skipbalance = false;
	if (n->left && n->right) {
		parent = NULL;
		n = n->right;
		while (n->left) {
			parent = n;
			n = n->left;
		}
		if (parent) balanceval = parent->val;
		else balanceval = n->val;
	} else {
		if (parent) balanceval = parent->val;
		else skipbalance = true;
	}
	remove(treenum,val);
	if (tree->root==NULL) return;
	tree->root->updateHeight();
	if (!skipbalance) balanceAvl(treenum,&tree->root,balanceval,false,false);
	drawAnim();
}

