/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include "treeanim.h"
#include "tree.h"

void TreeAnim::add(int treenum, int val, bool clear) {
	Tree *tree = forest->trees[treenum];
	if (tree->root==NULL) {
		tree->root = new Node(0,0);
		scaleSpacing(tree->root,NODEWIDTH,0);

		Tree *t = new Tree();
		t->root = new Node(val);
		forest->movables.push_back(t);
		dropToTree(t,treenum);

		forest->movables.pop_back();
		delete tree->root;
		tree->root = t->root;
		t->root = NULL;
		delete t;
		return;
	}

	Tree *t = new Tree();
	t->root = new Node(val);
	setRed(t->root);

	forest->movables.push_back(t);

	dropToTree(t,treenum,0,NODEWIDTH);

	Node **newpos = &tree->root;
	Node *parent = NULL;
	while (*newpos!=NULL) {
		if (parent!=NULL) clearMark(parent);
		parent = *newpos;
		markRed(parent);
		drawAnim();
		if (val<parent->val) newpos = &parent->left;
		else newpos = &parent->right;
	}
//	Node **newpos = tree->getParentLink(val);
//	if (*newpos!=NULL) return;

	*newpos = new Node(0,-NODEHEIGHT);
	scaleSpacing(*newpos,NODEWIDTH,0);

	jumpIntoTree(t,treenum);

	forest->movables.pop_back();
	delete *newpos;
	*newpos = t->root;
	t->root = NULL;
	delete t;

	growParent(*newpos);
	clearMark(*newpos);
	if (clear) clearMark(parent);
	drawAnim();

}

void TreeAnim::removeLeaf(Node **removepos, int treenum) {
	Node *n = *removepos;
	removeParent(n);
	*removepos = new Node(NODEWIDTH,0);
	Tree *t = new Tree();
	t->root = n;
	forest->movables.push_back(t);
	dropDown(t,treenum);
	forest->movables.pop_back();
	delete t;
	scaleSpacing(*removepos,0,-NODEHEIGHT);
	delete *removepos;
	*removepos = NULL;
}

void TreeAnim::removeSingleChildNode(Node **removepos, Node *parent, int treenum) {
	Node *n = *removepos;
	if (n->left==NULL && n->right==NULL) {
		removeLeaf(removepos,treenum);
		return;
	}
	removeParent(n,false);
	Node *child = n->right;
	float leftwidth = 0;
	if (n->left) {
		child = n->left;
		leftwidth = child->width;
	}
	removeParent(child,false);
	drawAnim();

	GLfloat t2pos[3];
	forest->getSubTreePos(treenum,child,t2pos);
	GLfloat endpos[3];
	forest->getSubTreePos(treenum,n,endpos);
	if (parent!=NULL) {
//		if (n==parent->left)
//			endpos[0] += n->width - child->width;
	}

	Tree *t1 = new Tree();
	t1->root = n;
	Tree *t2 = new Tree();
	memcpy(t2->pos,t2pos,sizeof(GLfloat)*3);
	t2->root = child;
	n->left = n->right = NULL;
	*removepos = new Node(n->width,n->height);
	forest->movables.push_back(t2);
	forest->movables.push_back(t1);

	dropDown(t1,treenum,leftwidth,0);
	delete t1;
	forest->movables.pop_back();

	moveTree(t2,endpos,false);
//	if (parent) setRelativeToNode(t2,treenum,parent);
	scaleSpacing(*removepos,child->width,child->height,false);
	drawAnim();
	delete *removepos;
	forest->movables.pop_back();
	*removepos = t2->root;
	t2->root = NULL;
	delete t2;
	growParent(child);
}

void TreeAnim::removeSimpleSubRoot(Node **removepos, int treenum) {
	Node *n = *removepos;
	Node *replacement = n->right;
	removeParent(n,false);
	removeParent(n->left,false);
	removeParent(n->right,false);
	drawAnim();

	GLfloat lpos[3], dpos[3], replacementpos[3];

	forest->getSubTreePos(treenum,n->left,lpos);
	forest->getNodePos(treenum,n,dpos);
	forest->getNodePos(treenum,replacement,replacementpos);

	Tree *ltree = new Tree();
	ltree->root = n->left;
	memcpy(ltree->pos,lpos,sizeof(GLfloat)*3);
	n->left = NULL;
	forest->movables.push_back(ltree);

	Tree *replacementtree = new Tree();
	replacementtree->root = replacement;
	memcpy(replacementtree->pos,replacementpos,sizeof(GLfloat)*3);
	n->right = NULL;
	forest->movables.push_back(replacementtree);

	Tree *t = new Tree();
	t->root = n;
	memcpy(t->pos,dpos,sizeof(GLfloat)*3);
	*removepos = new Node(n->width,n->height);
	forest->movables.push_back(t);
	dropDown(t,treenum,0,0,true);
	forest->movables.pop_back();
	delete t;

	moveTree(replacementtree,dpos,false);

	float nwidth = replacement->width+ltree->root->width;
	float nheight = replacement->height;
	if (ltree->root->height+NODEHEIGHT>nheight) nheight = ltree->root->height+NODEHEIGHT;

	scaleSpacing(*removepos,nwidth,nheight,false);
	drawAnim();
	forest->movables.pop_back();
	delete *removepos;
	*removepos = replacementtree->root;
	replacementtree->root = NULL;
	delete replacementtree;

	forest->movables.pop_back();
	replacement->left = ltree->root;
	ltree->root = NULL;
	delete ltree;

	growParent(replacement,false);
	growParent(replacement->left,false);
	clearMark(replacement);
	drawAnim();
}

void TreeAnim::copyRemove(Node **removepos, int treenum, Node *replacement, Node *parent) {
	Node *n = *removepos;
	GLfloat rpos[3], dpos[3];
	forest->getNodePos(treenum,n,dpos);
	forest->getNodePos(treenum,replacement,rpos);
//	dpos[2] += NODEWIDTH*0.001;

	Tree *t = new Tree();
	t->root = new Node(replacement->val);
	setRed(t->root);
	memcpy(t->pos,rpos,sizeof(GLfloat)*3);
	forest->movables.push_back(t);
	jumpTree(t,dpos);
	n->hide = true;

	clearMark(n);
	clearMark(t->root);
	drawAnim();

	if (parent==NULL) removeSingleChildNode(&n->right, n, treenum);
	else removeSingleChildNode(&parent->left, parent, treenum);

	forest->movables.pop_back();
	n->val = t->root->val;
	delete t;
	n->hide = false;
}

void TreeAnim::removeSubRoot(Node **removepos, int treenum) {
	Node *n = *removepos;
	Node **replacepos = &n->right;
	Node *parent = NULL;
	while (*replacepos!=NULL) {
		if (parent!=NULL) clearMark(parent);
		Node *cur = *replacepos;
		markRed(cur);
		drawAnim();
		if (cur->left==NULL) break;
		parent = *replacepos;
		replacepos = &parent->left;
	}


	Node *replacement = *replacepos;

	copyRemove(removepos,treenum,replacement,parent);
	return;

	if (n->right==replacement) {
		removeSimpleSubRoot(removepos,treenum);
		return;
	}

	removeParent(n,false);
	removeParent(n->left,false);
	removeParent(n->right,false);
	removeParent(replacement,false);
	removeParent(replacement->right,false);
	drawAnim();

	GLfloat lpos[3], rpos[3], dpos[3], childpos[3], replacementpos[3], childendpos[3];
	Node *child = replacement->left;
	if (replacement->right) child = replacement->right;

	forest->getSubTreePos(treenum,n->left,lpos);
	forest->getSubTreePos(treenum,n->right,rpos);
	forest->getNodePos(treenum,n,dpos);
	forest->getNodePos(treenum,replacement,replacementpos);
	if (child) {
		forest->getSubTreePos(treenum,child,childpos);
		forest->getSubTreePos(treenum,replacement,childendpos);
//		if (replacement==parent->left)
//			childendpos[0] += replacement->width - child->width;
	}


	Tree *childtree;
	if (child) {
		childtree = new Tree();
		childtree->root = child;
		memcpy(childtree->pos,childpos,sizeof(GLfloat)*3);
		replacement->left = replacement->right = NULL;
		forest->movables.push_back(childtree);
	}

	Tree *ltree = new Tree();
	ltree->root = n->left;
	memcpy(ltree->pos,lpos,sizeof(GLfloat)*3);
	n->left = NULL;
	forest->movables.push_back(ltree);

	Tree *rtree = new Tree();
	rtree->root = n->right;
	memcpy(rtree->pos,rpos,sizeof(GLfloat)*3);
	n->right = NULL;
	forest->movables.push_back(rtree);

	Tree *t = new Tree();
	t->root = n;
	memcpy(t->pos,dpos,sizeof(GLfloat)*3);
	*removepos = new Node(n->width,n->height);
	forest->movables.push_back(t);
	dropDown(t,treenum,0,0,true);
	forest->movables.pop_back();
	delete t;


	Tree *replacementtree = new Tree();
	replacementtree->root = replacement;
	memcpy(replacementtree->pos,replacementpos,sizeof(GLfloat)*3);
	*replacepos = new Node(replacement->width,replacement->height);

	forest->movables.push_back(replacementtree);
	moveTree(replacementtree,dpos);
	forest->movables.pop_back();


	delete *removepos;
	*removepos = replacementtree->root;
	replacementtree->root = NULL;
	delete replacementtree;
	forest->movables.pop_back();
	replacement->left = ltree->root;
	ltree->root = NULL;
	delete ltree;
	forest->movables.pop_back();
	replacement->right = rtree->root;
	rtree->root = NULL;
	delete rtree;

	growParent(replacement,false);
	growParent(replacement->left,false);
	growParent(replacement->right,false);
	clearMark(replacement);
	drawAnim();

	if (child) {
		moveTree(childtree,childendpos,false);
//		setRelativeToNode(childtree,treenum,parent);
		scaleSpacing(*replacepos,child->width,child->height,false);
		drawAnim();
		forest->movables.pop_back();
		delete *replacepos;
		*replacepos = childtree->root;
		childtree->root = NULL;
		delete childtree;
		growParent(child);
	} else {
		scaleSpacing(*replacepos,0,-NODEHEIGHT);
		delete *replacepos;
		*replacepos = NULL;
	}
}

void TreeAnim::remove(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	Node **removepos = &tree->root;
	Node *parent = NULL;
	while (*removepos!=NULL) {
		if (parent!=NULL) clearMark(parent);
		markRed(*removepos);
		drawAnim();
		if (val==(*removepos)->val) break;
		parent = *removepos;
		if (val<parent->val) removepos = &parent->left;
		else if (val>parent->val) removepos = &parent->right;
	}
	if (*removepos==NULL) {
		clearMark(parent);
		drawAnim();
		return;
	}
	Node *n = *removepos;
	if (n->left==NULL && n->right==NULL) {
		removeLeaf(removepos,treenum);
		return;
	}
	if (n->left==NULL || n->right==NULL) {
		removeSingleChildNode(removepos,parent,treenum);
		return;
	}

	removeSubRoot(removepos,treenum);

}



