/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include "treeanim.h"
#include "tree.h"
#include "vect.h"

void TreeAnim::percolateUp(Tree *tree, Tree *newtree, int npos, GLfloat *lastpos) {
	tree->root->updateWidth();
	setAbsolute(newtree);
	Tree *lastl = NULL, *lastr = NULL;
	while (npos>1) {
		int parentPos = npos/2;
		Node *parent = tree->heap[parentPos];
		markRed(parent);
		drawAnim();
		if (parent->val < newtree->root->val) break;
		removeParent(parent,false);
		if (parent->left) removeParent(parent->left,false);
		if (parent->right) removeParent(parent->right,false);
		drawAnim();


		Tree *l = NULL;
		if (2*parentPos!=npos) {
			l = new Tree();
			l->root = parent->left;
			forest->getTreePos(tree,l->pos);
			tree->addHeapSubTreePos(2*parentPos,l->pos);
			forest->movables.push_back(l);
		}
		Tree *r = NULL;
		if (2*parentPos+1!=npos && 2*parentPos+1< int(tree->heap.size())) {
			r = new Tree();
			r->root = parent->right;
			forest->getTreePos(tree,r->pos);
			tree->addHeapSubTreePos(2*parentPos+1,r->pos);
			forest->movables.push_back(r);
		}
		Tree *c = new Tree();
		c->root = parent;
		forest->getTreePos(tree,c->pos);
		tree->addHeapPos(parentPos,c->pos);
		forest->movables.push_back(c);
		GLfloat curpos[3];
		Vect_copy3f(curpos,c->pos);

		Node **cpos = tree->getHeapLink(npos);
		delete *cpos;
		*cpos = NULL;

		parent->left = NULL;
		parent->right = NULL;

		Node **ppos = tree->getHeapLink(parentPos);
		*ppos = new Node(parent->width,parent->height);

		moveTree(c,lastpos,true);
		tree->heap[npos] = c->root;

		forest->movables.pop_back();
		if (l) forest->movables.pop_back();
		if (r) forest->movables.pop_back();

		if (lastl) {
			forest->movables.pop_back();
			c->root->left = lastl->root;
			c->pos[0] -= lastl->root->width;
			lastl->root = NULL;
			delete lastl;
			lastl = NULL;
		}
		if (lastr) {
			forest->movables.pop_back();
			c->root->right = lastr->root;
			lastr->root = NULL;
			delete lastr;
			lastr = NULL;
		}


		if (npos==2*parentPos) {
			// l == NULL
			lastr = r;
			lastl = c;
			if (lastl) forest->movables.push_back(lastl);
			if (lastr) forest->movables.push_back(lastr);
		} else {
			// r == NULL
			lastl = l;
			lastr = c;
			if (lastl) forest->movables.push_back(lastl);
			if (lastr) forest->movables.push_back(lastr);
		}

		if (c->root->left) growParent(c->root->left,false);
		if (c->root->right) growParent(c->root->right,false);
		clearMark(c->root);
		drawAnim();


		Vect_copy3f(lastpos,curpos);
		npos = parentPos;
	}
	int parentPos = npos/2;
	Node *parent = tree->heap[parentPos];

	if (parentPos>0) {
		clearMark(parent);
		drawAnim();
	}

	jumpTree(newtree,lastpos);

	Node **cpos = tree->getHeapLink(npos);
	delete *cpos;
	*cpos = newtree->root;
	tree->heap[npos] = newtree->root;

	if (lastl) {
		forest->movables.pop_back();
		newtree->root->left = lastl->root;
		lastl->root = NULL;
		delete lastl;
		lastl = NULL;
	}
	if (lastr) {
		forest->movables.pop_back();
		newtree->root->right = lastr->root;
		lastr->root = NULL;
		delete lastr;
		lastr = NULL;
	}

	if (newtree->root->left) growParent(newtree->root->left,false);
	if (newtree->root->right) growParent(newtree->root->right,false);
	growParent(newtree->root,false);
	clearMark(newtree->root);

	newtree->root = NULL;
	delete newtree;
	forest->movables.pop_back();

	drawAnim();

}

void TreeAnim::addHeap(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	if (tree->heap.size()==0) tree->heap.push_back(NULL);

	if (tree->root==NULL) {
		add(treenum,val);
		tree->heap.push_back(tree->root);
		return;
	}

	Tree *t = new Tree();
	t->root = new Node(val);
	setRed(t->root);
	forest->movables.push_back(t);
	dropToTree(t,treenum,0,NODEWIDTH);

	int npos = tree->heap.size();
	Node **nptr = tree->getHeapLink(npos);
	*nptr = new Node(0,-NODEHEIGHT);
	scaleSpacing(*nptr,NODEWIDTH,0,true);
	GLfloat pos[3];
	forest->getTreePos(tree,pos);
	tree->addHeapPos(npos,pos);
	tree->heap.push_back(NULL);

	percolateUp(tree,t,npos,pos);
}

void TreeAnim::percolateDown(Tree *tree, Tree *t, int npos, GLfloat *lastpos, Tree *l, Tree *r) {
	while (true) {
		int child1pos = 2*npos;
		int child2pos = 2*npos+1;
		if (child1pos>=int(tree->heap.size())) break;

		Node *child1 = tree->heap[child1pos];
		Node *child2 = NULL;
		markRed(child1);
		if (child2pos<int(tree->heap.size())) {
			child2 = tree->heap[child2pos];
			markRed(child2);
		}
		drawAnim();

		int minchild = child1pos;
		if (child2 && child2->val<child1->val) {
			clearMark(child1);
			drawAnim();
			minchild = child2pos;
		} else {
			if (child2) {
				clearMark(child2);
				drawAnim();
			}
		}
		if (t->root->val<tree->heap[minchild]->val) {
			clearMark(tree->heap[minchild]);
			drawAnim();
			break;
		}

		Node *mc = tree->heap[minchild];

		if (mc->left) removeParent(mc->left,false);
		if (mc->right) removeParent(mc->right,false);
		drawAnim();

		GLfloat treepos[3], startpos[3];
		if (minchild==child1pos) {
			Vect_copy3f(treepos,l->pos);

			forest->movables.pop_back();
			forest->movables.pop_back();
			if (r) {
				forest->movables.pop_back();
				forest->movables.push_back(r);
			}
			forest->movables.push_back(t);
		} else {
			Vect_copy3f(treepos,r->pos);

			forest->movables.pop_back();
			forest->movables.pop_back();
			forest->movables.pop_back();
			forest->movables.push_back(l);
			forest->movables.push_back(t);
		}
		Vect_copy3f(startpos,treepos);
		tree->addHeapPos(minchild,startpos);

		mc->updateWidth();
		mc->updateHeight();
		float mcwidth = mc->width;
		float mcheight = mc->height;

		Tree *c = new Tree();
		c->root = mc;
		Vect_copy3f(c->pos,startpos);



		Tree *nr = NULL;
		if (mc->right) {
			nr = new Tree();
			nr->root = mc->right;
			mc->right = NULL;
			Vect_copy3f(nr->pos,treepos);
			tree->addHeapSubTreePos(2*minchild+1,nr->pos);
			forest->movables.push_back(nr);
		}
		Tree *nl = NULL;
		if (mc->left) {
			nl = new Tree();
			nl->root = mc->left;
			mc->left = NULL;
			Vect_copy3f(nl->pos,treepos);
			tree->addHeapSubTreePos(2*minchild,nl->pos);
			forest->movables.push_back(nl);
		}

		forest->movables.push_back(c);
		moveTree(c,lastpos,true);

		tree->heap[npos] = mc;
		tree->heap[minchild] = NULL;
		Node **root = tree->getHeapLink(npos);
		delete *root;
		*root = mc;
		c->root = NULL;
		forest->movables.pop_back();
		delete c;

		if (nl) forest->movables.pop_back();
		if (nr) forest->movables.pop_back();
		forest->movables.pop_back();
		if (r) forest->movables.pop_back();

		if (minchild==child1pos) {
			if (r) {
				mc->right = r->root;
				r->root = NULL;
				delete r;
				growParent(mc->right,false);
			} else mc->right = NULL;
			mc->left = new Node(mcwidth,mcheight);
			l->root = NULL;
			delete l;
		} else {
			mc->left = l->root;
			l->root = NULL;
			delete l;
			if (r) {
				r->root = NULL;
				delete r;
			}
			growParent(mc->left,false);
			mc->right = new Node(mcwidth,mcheight);
		}

		if (nl) forest->movables.push_back(nl);
		if (nr) forest->movables.push_back(nr);
		forest->movables.push_back(t);

		clearMark(mc);
		growParent(mc,false);
		drawAnim();


		l = nl;
		r = nr;
		Vect_copy3f(lastpos,startpos);
		npos = minchild;
	}

	jumpTree(t,lastpos,false);
	clearMark(t->root);
	drawAnim();
	if (l) {
		forest->movables.pop_back();
		t->root->left = l->root;
		l->root = NULL;
		delete l;
		growParent(t->root->left,false);
	}
	if (r) {
		forest->movables.pop_back();
		t->root->right = r->root;
		r->root = NULL;
		delete r;
		growParent(t->root->right,false);
	}
	forest->movables.pop_back();
	Node **root = tree->getHeapLink(npos);
	delete *root;
	*root = t->root;
	tree->heap[npos] = t->root;
	t->root = NULL;
	delete t;
	growParent(*root,false);
	drawAnim();
}

void TreeAnim::removeHeap(int treenum) {
	Tree *tree = forest->trees[treenum];
	if (tree->root->left) removeParent(tree->root->left,false);
	if (tree->root->right) removeParent(tree->root->right,false);
	drawAnim();

	Tree *c = new Tree();
	forest->getTreePos(treenum,c->pos);
	tree->addHeapPos(1,c->pos);
	c->root = tree->root;

	GLfloat lastpos[3];
	if (tree->heap.size()>2) {
		forest->getTreePos(treenum,lastpos);
		tree->addHeapPos(tree->heap.size()-1,lastpos);
	}
	GLfloat toppos[3];
	Vect_copy3f(toppos,c->pos);
	toppos[1] += NODEWIDTH;

	
	Node **spacing = tree->getHeapLink(tree->heap.size()-1);
	(*spacing)->spacing = true;
	(*spacing)->height = 0;
	float h1 = tree->root->updateHeight();
	(*spacing)->height = -NODEHEIGHT;
	float h2 = tree->root->updateHeight();
	(*spacing)->spacing = false;
	(*spacing)->height = 0;
	tree->root->updateHeight();


	Tree *r = NULL;
	if (tree->root->right) {
		r = new Tree();
		forest->getTreePos(treenum,r->pos);
		tree->addHeapSubTreePos(3,r->pos);
		forest->movables.push_back(r);
		r->root = tree->root->right;
		tree->root->right = NULL;
	}
	Tree *l = NULL;
	if (tree->root->left) {
		l = new Tree();
		forest->getTreePos(treenum,l->pos);
		tree->addHeapSubTreePos(2,l->pos);
		forest->movables.push_back(l);
		l->root = tree->root->left;
		tree->root->left = NULL;
	}

	tree->root = new Node(tree->root->width,tree->root->height);

	forest->movables.push_back(c);
	dropDown(c,treenum,0,0,true);
	forest->movables.pop_back();
	delete c;
	tree->heap[1] = NULL;

	if (l==NULL) {
		delete tree->root;
		tree->root = NULL;
		return;
	}

	Node *last = tree->heap[tree->heap.size()-1];
	tree->heap.pop_back();
	removeParent(last,true);

	if (tree->heap.size()==3) {
		spacing = &r->root;
	} else if (tree->heap.size()==2) {
		spacing = &l->root;
	} else {
		spacing = tree->getHeapLink(tree->heap.size());
	}
	*spacing = new Node(NODEWIDTH,0);
	Tree *lasttree = new Tree();
	lasttree->root = last;
	Vect_copy3f(lasttree->pos,lastpos);
	forest->movables.push_back(lasttree);

	int index = tree->heap.size();
	while (index>3) index /= 2;
	if (index==2) {
		toppos[0] -= NODEWIDTH;
		GLfloat pos[3];
		if (r) {
			Vect_copy3f(pos,r->pos);
			pos[0] -= NODEWIDTH;
			moveTree(r,pos,false);
		}
	}

	scaleSpacing(tree->root,tree->root->width-NODEWIDTH,tree->root->height+h2-h1,false);
	scaleSpacing(*spacing,0,-NODEHEIGHT,false);
	jumpTree(lasttree,toppos,false);
	markRed(lasttree->root);
	drawAnim();
	delete *spacing;
	*spacing = NULL;
	toppos[1] -= NODEWIDTH;
	percolateDown(tree,lasttree,1,toppos,l,r);
}

void TreeAnim::buildHeap(int treenum, int *values, int nvalues) {
	Tree *tree = forest->trees[treenum];
	if (tree->heap.size()==0) tree->heap.push_back(NULL);
	for (int i = 0; i < nvalues; i++) {
		int npos = tree->heap.size();
		Node **nptr = tree->getHeapLink(npos);
		*nptr = new Node(0,-NODEHEIGHT);
		scaleSpacing(*nptr,NODEWIDTH,0,true);

		GLfloat startpos[3], destpos[3];
		forest->getTreePos(treenum,startpos);
		Vect_copy3f(destpos,startpos);
		if (tree->root!=NULL)
			tree->addPos(tree->root,startpos);
		startpos[1] = forest->getTopY()+NODEWIDTH;

		tree->addHeapPos(npos,destpos);
		Tree *t = new Tree();
		t->root = new Node(values[i]);
		forest->movables.push_back(t);

		Vect_copy3f(t->pos,startpos);
		jumpTree(t,destpos,true);

		forest->movables.pop_back();
		delete *nptr;
		*nptr = t->root;
		t->root = NULL;
		delete t;
		tree->heap.push_back(*nptr);
		growParent(*nptr,true);
	}

	for (int i = nvalues/2; i > 0; i--) {
		Node *n = tree->heap[i];
		if (n->left) removeParent(n->left,false);
		if (n->right) removeParent(n->right,false);
		removeParent(n,false);
		drawAnim();


		GLfloat toppos[3];
		forest->getTreePos(treenum,toppos);
		tree->addHeapPos(1,toppos);
		toppos[1] += NODEWIDTH;


		Tree *c = new Tree();
		forest->getTreePos(treenum,c->pos);
		tree->addHeapPos(i,c->pos);
		c->root = n;

		GLfloat holepos[3];
		Vect_copy3f(holepos,c->pos);
	
		Tree *r = NULL;
		if (n->right) {
			r = new Tree();
			forest->getTreePos(treenum,r->pos);
			tree->addHeapSubTreePos(2*i+1,r->pos);
			forest->movables.push_back(r);
			r->root = n->right;
			n->right = NULL;
		}
		Tree *l = NULL;
		if (n->left) {
			l = new Tree();
			forest->getTreePos(treenum,l->pos);
			tree->addHeapSubTreePos(2*i,l->pos);
			forest->movables.push_back(l);
			l->root = n->left;
			n->left = NULL;
		}

		Node **nlink = tree->getHeapLink(i);
		*nlink = new Node(c->root->width,c->root->height);
		tree->heap[i] = NULL;

		forest->movables.push_back(c);
		jumpTree(c,toppos,false);
		markRed(c->root);
		drawAnim();
		percolateDown(tree,c,i,holepos,l,r);

	}
}

