/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include <stack>
#include "treeanim.h"
#include "tree.h"

using std::stack;

/*
static bool checkNode(int dist, Node *n, stack<int> &dists) {
	if (n==NULL) {
		dists.push(dist);
		return true;
	}
	if (!n->red) dist++;
	bool ok = true;
	if (!checkNode(dist,n->left,dists)) ok = false;
	if (!checkNode(dist,n->right,dists)) ok = false;
	if (n->red) {
		if (n->left && n->left->red) {
			printf("invariant failed, red node %d has red child %d\n",n->val,n->left->val);
			ok = false;
		}
		if (n->right && n->right->red) {
			printf("invariant failed, red node %d has red child %d\n",n->val,n->right->val);
			ok = false;
		}
	}
	return ok;
}

static void checkRedBlackInvariant(TreeAnim *ta, Tree *t) {
	stack<int> dists;
	bool ok = checkNode(0,t->root,dists);
	if (t->root && t->root->red) {
		printf("invariant failed, root is red\n");
		ok = false;
	}
	int dist = dists.top();
	while (!dists.empty()) {
		if (dists.top()!=dist) {
			printf("invariant failed, number of black nodes on path differs\n");
			ok = false;
			break;
		}
		dists.pop();
	}
//	if (ok) printf("invariant ok\n");
	if (!ok) ta->drawAnim(10000);
}
*/

void TreeAnim::reorient(Tree *t, Node *current, Node *parent, Node *grand, Node **grandpos) {
	if (!current->red) {
		if (current->left) markBlack(current->left);
		if (current->right) markBlack(current->right);
		markRed(current);
		drawAnim();
	}

	if (parent && parent->red) {
		// grand can't be NULL, because the root is always black
		if (parent==grand->left) {
			if (current==parent->left) {
				rotateRight(t,grandpos,false);
				markBlack(parent);
				markRed(grand);
				drawAnim();
			} else {
				rotateLeft(t,&grand->left,false);
				rotateRight(t,grandpos,false);
				markBlack(current);
				markRed(grand);
				drawAnim();
			}
		} else {
			if (current==parent->right) {
				rotateLeft(t,grandpos,false);
				markBlack(parent);
				markRed(grand);
				drawAnim();
			} else {
				rotateRight(t,&grand->right,false);
				rotateLeft(t,grandpos,false);
				markBlack(current);
				markRed(grand);
				drawAnim();
			}
		}
	}

	if (t->root->red) {
		markBlack(t->root);
		drawAnim();
	}
}

void TreeAnim::addRedBlack(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	if (tree->root==NULL) {
		tree->root = new Node(0,0);
		scaleSpacing(tree->root,NODEWIDTH,0);

		Tree *t = new Tree();
		t->root = new Node(val);
		setBlack(t->root);
		forest->movables.push_back(t);
		dropToTree(t,treenum);

		forest->movables.pop_back();
		delete tree->root;
		tree->root = t->root;
		t->root = NULL;
		delete t;
		return;
	}

	Tree *t = new Tree();
	t->root = new Node(val);
	setRed(t->root);
	setGreen(t->root);
	t->root->red = true;

	forest->movables.push_back(t);

	dropToTree(t,treenum,0,NODEWIDTH);
	setAbsolute(t);

	Node **newpos = &tree->root;
	Node *current = tree->root;
	Node **parentpos = NULL;
	Node *parent = NULL;
	Node **grandpos = NULL;
	Node *grand = NULL;

	while (current!=NULL) {
		if (parent!=NULL) {
			if (parent->red) markRed(parent);
			else markBlack(parent);
		}

		markGreen(current);
		drawAnim();

		if (current->left && current->right) {
			if (current->left->red && current->right->red) {
				reorient(tree,current,parent,grand,grandpos);
			}
		}

		grandpos = parentpos;
		parentpos = newpos;

		grand = parent;
		parent = current;

		if (val<current->val) newpos = &current->left;
		else newpos = &current->right;
		current = *newpos;

	}
//	Node **newpos = tree->getParentLink(val);
//	if (*newpos!=NULL) return;

	*newpos = new Node(0,-NODEHEIGHT);
	setRelativeToRoot(t,treenum);
	scaleSpacing(*newpos,NODEWIDTH,0);

	jumpIntoTree(t,treenum);

	forest->movables.pop_back();
	delete *newpos;
	*newpos = t->root;
	t->root = NULL;
	delete t;

	growParent(*newpos);
	markRed(*newpos);
	if (parent->red) markRed(parent);
	else markBlack(parent);
	drawAnim();

	if (parent->red)
		reorient(tree,*newpos,parent,grand,grandpos);

//	checkRedBlackInvariant(this,tree);
}

void TreeAnim::removeRedBlack(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	markRed(tree->root);
	drawAnim();
	Node **currentpos = &tree->root;
	Node *current = tree->root;
	Node *parent = NULL;
	Node **parentpos = NULL;
	Node *sibling = NULL;
	bool searchvalue = true;
	bool searchleft = true;
	int removeval = -1;
	stack<Node*> removed;
	while (true) {
		bool rightnode = false;
		if (searchvalue) {
			if (current->val==val) rightnode = true;
		} else {
			if (searchleft) {
				if (current->left==NULL) rightnode = true;
			} else {
				if (current->right==NULL) rightnode = true;
			}
		}
		markGreen(current);
		drawAnim();
		if (rightnode) {
			if (!searchvalue) {
				Node *removable = *(tree->getParentLink(removeval));
				Tree *t = new Tree();
				t->root = new Node(current->val);
				if (current->red) setRed(t->root);
				else setBlack(t->root);
				setGreen(t->root);
				GLfloat end[3];
				forest->getNodePos(treenum,removable,end);
				forest->getNodePos(treenum,current,t->pos);
				forest->movables.push_back(t);
				if (removable->red) {
					markRed(t->root);
//					markRed(removable);
				} else {
					markBlack(t->root);
//					markBlack(removable);
				}
				jumpTree(t,end);
				removable->hide = true;
				removed.push(removable);
				setRelativeToNode(t,treenum,removable);
			}
			if (current->left==NULL && current->right==NULL) {
				removeLeaf(currentpos,treenum);
				break;
			} else {
				removeval = current->val;
				searchvalue = false;
				parentpos = currentpos;
				parent = current;
				if (current->right) {
					searchleft = true;
					sibling = current->left;
					currentpos = &current->right;
					current = current->right;
				} else {
					searchleft = false;
					sibling = current->right;
					currentpos = &current->left;
					current = current->left;
				}
			}
		} else {
			parentpos = currentpos;
			parent = current;
			if (searchvalue) {
				if (val<current->val) {
					sibling = current->right;
					currentpos = &current->left;
					current = current->left;
				} else {
					sibling = current->left;
					currentpos = &current->right;
					current = current->right;
				}
			} else {
				if (searchleft) {
					sibling = current->right;
					currentpos = &current->left;
					current = current->left;
				} else {
					sibling = current->left;
					currentpos = &current->right;
					current = current->right;
				}
			}
		}

		if (!current->red && sibling && sibling->red) {
			// parent==root
			if (sibling==parent->right) rotateLeft(tree,parentpos,false);
			else rotateRight(tree,parentpos,false);
			markGreenRed(parent);
			markBlack(sibling);
			drawAnim();
			if (current==parent->left) {
				parentpos = &sibling->left;
				sibling = parent->right;
			} else {
				parentpos = &sibling->right;
				sibling = parent->left;
			}
		}
		if (current->red) {
			// parent==root
			markGreenBlack(parent);
			drawAnim();
		} else if ((!current->left || !current->left->red) && (!current->right || !current->right->red)) {
			// both children black
			if (!sibling || ((!sibling->left || !sibling->left->red) && (!sibling->right || !sibling->right->red))) {
				// both sibling's children black
				if (sibling) markRed(sibling);
				markRed(current);
				markGreenBlack(parent);
				drawAnim();
			} else {
				if (sibling==parent->right) {
					if (sibling->right && sibling->right->red) {
						rotateLeft(tree,parentpos,false);
						markGreenBlack(parent);
						markRed(current);
						if (tree->root==sibling) markBlack(sibling);
						else markRed(sibling);
						markBlack(sibling->right);
						drawAnim();
					} else {
						Node *newroot = sibling->left;
						rotateRight(tree,&parent->right,false);
						rotateLeft(tree,parentpos,false);
						markGreenBlack(parent);
						markRed(current);
						markBlack(sibling);
						if (tree->root==newroot) markBlack(newroot);
						else markRed(newroot);
						drawAnim();
					}
				} else {
					if (sibling->left && sibling->left->red) {
						rotateRight(tree,parentpos,false);
						markGreenBlack(parent);
						markRed(current);
						if (tree->root==sibling) markBlack(sibling);
						else markRed(sibling);
						markBlack(sibling->left);
						drawAnim();
					} else {
						Node *newroot = sibling->right;
						rotateLeft(tree,&parent->left,false);
						rotateRight(tree,parentpos,false);
						markGreenBlack(parent);
						markRed(current);
						markBlack(sibling);
						if (tree->root==newroot) markBlack(newroot);
						else markRed(newroot);
						drawAnim();
					}
				}
			}
		} else {
			if (tree->root->red) {
				if (parent==tree->root) markGreenBlack(tree->root);
				else markBlack(tree->root);
				drawAnim();
			}
		}

		if (removeval!=parent->val) {
			if (parent->red) markRed(parent);
			else markBlack(parent);
		}
//		checkRedBlackInvariant(this,tree);
	}
	drawAnim();
	while (!removed.empty()) {
		Tree *t = forest->movables.back();
		forest->movables.pop_back();
		Node *removable = removed.top();
		removed.pop();
		removable->val = t->root->val;
		removable->hide = false;
		if (removable->red) setRed(removable);
		else setBlack(removable);
		delete t;
	}
//	checkRedBlackInvariant(this,tree);
}


