/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL_opengl.h>
#include <stack>
#include "treeanim.h"
#include "tree.h"
#include "vect.h"

using std::stack;

float TreeAnim::scaleSpacingRotLeft(Node *spacing, Node *n, Tree *t, float minheight) {
	if (spacing==NULL) return -1;
	n->updateHeight();
	float origheight = n->height;
	float nheight1 = 0;
	float nheight2 = 0;
	float nheight3 = 0;
	if (n->right->right) nheight1 = n->right->right->height+NODEHEIGHT;
	if (n->right->left) nheight2 = n->right->left->height+2*NODEHEIGHT;
	if (n->left) nheight3 = n->left->height+2*NODEHEIGHT;
	float height = nheight1;
	if (nheight2>height) height = nheight2;
	if (nheight3>height) height = nheight3;
	float ctotheight = t->root->updateHeight();
	n->spacing = true;
	n->height = height;
	float ntotheight = t->root->updateHeight();
	n->spacing = false;
	n->height = origheight;
	t->root->updateHeight();
	float nheight = ntotheight;
	if (nheight<minheight) nheight = minheight;
	if (nheight<ctotheight) return nheight;
	scaleSpacing(spacing,spacing->width,nheight,false);
	return -1;
}

float TreeAnim::scaleSpacingRotRight(Node *spacing, Node *n, Tree *t, float minheight) {
	if (spacing==NULL) return -1;
	n->updateHeight();
	float origheight = n->height;
	float nheight1 = 0;
	float nheight2 = 0;
	float nheight3 = 0;
	if (n->left->left) nheight1 = n->left->left->height+NODEHEIGHT;
	if (n->left->right) nheight2 = n->left->right->height+2*NODEHEIGHT;
	if (n->right) nheight3 = n->right->height+2*NODEHEIGHT;
	float height = nheight1;
	if (nheight2>height) height = nheight2;
	if (nheight3>height) height = nheight3;
	float ctotheight = t->root->updateHeight();
	n->spacing = true;
	n->height = height;
	float ntotheight = t->root->updateHeight();
	n->spacing = false;
	n->height = origheight;
	float nheight = ntotheight;
	if (nheight<minheight) nheight = minheight;
	if (nheight<ctotheight) return nheight;
	scaleSpacing(spacing,spacing->width,nheight,false);
	return -1;
}

void TreeAnim::splay(Tree *tree, Node **root, int val, bool mark, Node *spacing, float minheight) {
	stack<Node*> nstack;
	Node *n = *root;
	while (n->val!=val) {
		if (mark) {
			markRed(n);
			drawAnim();
		}
		nstack.push(n);
		if (mark) clearMark(n);
		if (val<n->val) {
			n = n->left;
		} else {
			n = n->right;
		}
	}
	if (mark) {
		markRed(n);
		drawAnim();
		clearMark(n);
	}
	bool first = true;
	while (nstack.size()>=2) {
		Node *child = n;
		Node *parent = nstack.top();
		nstack.pop();
		Node *gparent = nstack.top();
		nstack.pop();

		if (!mark) {
			if (!first) {
				markRed(parent);
				drawAnim();
			}
			first = false;
		} else {
			markRed(parent);
			drawAnim();
		}
		clearMark(parent);
		markRed(gparent);
		drawAnim();

		Node **rpos = root;
		if (nstack.size()>0) {
			if (nstack.top()->left==gparent)
				rpos = &nstack.top()->left;
			else
				rpos = &nstack.top()->right;
		}

		bool left = parent->left==child;
		bool gleft = gparent->left==parent;
		if (left && !gleft) {
			float h = scaleSpacingRotRight(spacing,gparent->right,tree,minheight);
			rotateRight(tree,&gparent->right);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
			h = scaleSpacingRotLeft(spacing,*rpos,tree,minheight);
			rotateLeft(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		} else if (!left && gleft) {
			float h = scaleSpacingRotLeft(spacing,gparent->left,tree,minheight);
			rotateLeft(tree,&gparent->left);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
			h = scaleSpacingRotRight(spacing,*rpos,tree,minheight);
			rotateRight(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		} else if (left && gleft) {
			float h = scaleSpacingRotRight(spacing,*rpos,tree,minheight);
			rotateRight(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
			h = scaleSpacingRotRight(spacing,*rpos,tree,minheight);
			rotateRight(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		} else {
			float h = scaleSpacingRotLeft(spacing,*rpos,tree,minheight);
			rotateLeft(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
			h = scaleSpacingRotLeft(spacing,*rpos,tree,minheight);
			rotateLeft(tree,rpos);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		}
	}
	if (nstack.size()>0) {
		Node *child = n;
		Node *parent = nstack.top();
		nstack.pop();
		if (!mark) {
			if (!first) {
				markRed(parent);
				drawAnim();
			}
			first = false;
		} else {
			markRed(parent);
			drawAnim();
		}
		if (parent->left==child) {
			float h = scaleSpacingRotRight(spacing,*root,tree,minheight);
			rotateRight(tree,root);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		} else {
			float h = scaleSpacingRotLeft(spacing,*root,tree,minheight);
			rotateLeft(tree,root);
			if (h>=0) scaleSpacing(spacing,spacing->width,h,true);
		}
	}
	drawAnim();
}

void TreeAnim::addSplay(int treenum, int val) {
	add(treenum,val,false);
	Tree *tree = forest->trees[treenum];
	splay(tree,&tree->root,val,false);
}

void TreeAnim::removeSplay(int treenum, int val) {
	Tree *tree = forest->trees[treenum];
	splay(tree,&tree->root,val,true);

	if (tree->root->left) removeParent(tree->root->left,false);
	if (tree->root->right) removeParent(tree->root->right,false);
	drawAnim();

	Tree *l = NULL;
	if (tree->root->left){
		l = new Tree();
		forest->getSubTreePos(treenum,tree->root->left,l->pos);
		l->root = tree->root->left;
		forest->movables.push_back(l);
	}
	Tree *r = NULL;
	if (tree->root->right) {
		r = new Tree();
		forest->getSubTreePos(treenum,tree->root->right,r->pos);
		r->root = tree->root->right;
		forest->movables.push_back(r);
	}
	Tree *n = new Tree();
	forest->getNodePos(treenum,tree->root,n->pos);
	n->root = tree->root;
	n->root->left = NULL;
	n->root->right = NULL;
	tree->root = new Node(n->root->width,n->root->height);
	forest->movables.push_back(n);
	dropDown(n,treenum,0,0,true);
	forest->movables.pop_back();
	delete n;

	if (l==NULL) {
		if (r==NULL) {
			delete tree->root;
			tree->root = NULL;
			return;
		}
		GLfloat pos[3];
		Vect_copy3f(pos,r->pos);
		pos[0] -= NODEWIDTH;
		pos[1] += NODEHEIGHT;
		moveTree(r,pos,false);
		scaleSpacing(tree->root,r->root->width,r->root->height,false);
		drawAnim();
		tree->root = r->root;
		r->root = NULL;
		delete r;
		forest->movables.pop_back();
		return;
	}

	GLfloat pos[3];
	Vect_copy3f(pos,l->pos);
	pos[1] += NODEHEIGHT;
	moveTree(l,pos,false);
	if (r!=NULL) {
		Vect_copy3f(pos,r->pos);
		pos[0] -= NODEWIDTH;
		moveTree(r,pos,false);
	}
	float nwidth = l->root->width;
	float nheight = l->root->height;
	if (r) {
		nwidth += r->root->width;
		if (r->root->height+NODEHEIGHT>nheight) nheight = r->root->height+NODEHEIGHT;
	}
	scaleSpacing(tree->root,nwidth,nheight,false);
	drawAnim();

	Node *node = l->root;
	while (node->right!=NULL)
		node = node->right;
	float minheight = 0;
	if (r) minheight = r->root->height+NODEHEIGHT;
	splay(l,&l->root,node->val,true,tree->root,minheight);

	if (r) {
		l->root->right = r->root;
		r->root = NULL;
		delete r;
		forest->movables.pop_back();
		growParent(l->root->right);
	}
	delete tree->root;
	tree->root = l->root;
	l->root = NULL;
	delete l;
	forest->movables.pop_back();
}

