/*
    Data structures & Algorithms
    Copyright 2004, 2005 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#ifndef __VECT_H
#define __VECT_H

class Vect3 {
private:
	GLfloat vector[3];
public:
	Vect3();
	Vect3(const GLfloat *v);
	Vect3(GLfloat x, GLfloat y, GLfloat z);
	Vect3(const Vect3 &v);
	Vect3 &operator+=(const Vect3 &v);
	Vect3 operator+(const Vect3 &v) const;
	Vect3 &operator-=(const Vect3 &v);
	Vect3 operator-(const Vect3 &v) const;
	Vect3 &operator*=(GLfloat f);
	Vect3 operator*(GLfloat f) const;
	Vect3 &operator/=(GLfloat f);
	Vect3 operator/(GLfloat f) const;
	GLfloat abs() const;
	Vect3 &normalize();
	void get(GLfloat *v) const;
	static GLfloat dot(const Vect3 &v1, const Vect3 &v2);
	static Vect3 cross(const Vect3 &v1, const Vect3 &v2);
};
Vect3 operator*(GLfloat f, const Vect3 &v);

void Vect_add3f(GLfloat *v1, const GLfloat *v2);
void Vect_sub3f(GLfloat *v1, const GLfloat *v2);
void Vect_normalize3f(GLfloat *v);
void Vect_copy3f(GLfloat *out, const GLfloat *in);
void Vect_cross3f(GLfloat *out, const GLfloat *v1, const GLfloat *v2);
float Vect_dot3f(const GLfloat *v1, const GLfloat *v2);
void Vect_mult3f(GLfloat *v, float f);
float Vect_abs3f(const GLfloat *v);


void printVect3f(GLfloat *vector);
void printVect4f(GLfloat *vector);


#endif
